"""
ARCTL CLI with trace explanation.
"""

import cmd
import argparse

from pynusmv.init import init_nusmv, deinit_nusmv, reset_nusmv
from pynusmv.fsm import BddFsm
from pynusmv.exception import PyNuSMVError

from pynusmv_tools.arctl.parsing import parseArctl
from pynusmv_tools.arctl.check import checkArctl

from pyparsing import ParseException


def print_bdd(bdd):
    """Print bdd at stdout."""
    for var, val in bdd.get_str_values().items():
        print(var, "=", val)
        
def print_path(path, inputs, loop):
    c = 0
    # print first
    if loop is not None and loop == path[0]:
        print("----- Loop starts here")
    print("----- State", str(c))
    c += 1
    print_bdd(path[0])
    # print path
    for i, s in zip(path[1::2], path[2::2]):
        print("----- Inputs")
        print_bdd(i)
        if loop is not None and loop == s:
            print("----- Loop starts here")
        print("----- State", str(c))
        c += 1
        print_bdd(s)
    # print loop
    if loop is not None:
        print("----- Inputs")
        print_bdd(inputs)
        print("----- State", str(c))
        c += 1
        print_bdd(loop)


class ARCTLshell(cmd.Cmd):

    prompt = "> "
    
    fsm = None
    fsmpath = None
    
    def preloop(self):
        init_nusmv()
        
    def postloop(self):
        deinit_nusmv()
        
        
    def do_read(self, arg):
        """
        Read file as SMV model to get FSM.
        usage: read FILEPATH
        """
        if len(arg) < 1:
            print("[ERROR] read command needs the SMV model path.")
        else:
            if self.fsm is not None:
                reset_nusmv()
            try:
                self.fsm = BddFsm.from_filename(arg)
                self.fsmpath = arg
            except Exception as e:
                print("[ERROR]", e)
            
            
    def do_check(self, arg):
        """
        Check an ARCTL specification on the current FSM.
        usage: check SPEC
        """
        if self.fsm is None:
            print("[ERROR] No FSM read. Use read command before.")
        elif arg == "":
            print("[ERROR] Need a specification.")
        else:
            try:
                spec = parseArctl(arg)[0]
                (res, (wit, (inp, loop))) = checkArctl(self.fsm, spec)
                if res:
                    print("The specification", arg, "is true,",
                          "witnessed by")
                    #print_path(wit, inp, loop)
                        
                else:
                    print("The specification", arg, "is false,",
                          "as shown by")
                    #print_path(wit, inp, loop)
                    
            except PyNuSMVError as e:
                print("[ERROR]", e)
                
            except ParseException as e:
                print("[ERROR]", e)
                
                
    def do_fsm(self, arg):
        """
        Print the path to the currently read FSM.
        usage: fsm
        """
        if self.fsmpath is None:
            print("No FSM.")
        else:
            print(self.fsmpath)
        
            
            
    def do_quit(self, arg):
        """
        Quit this prompt.
        usage: quit
        """
        return True
        
        
def main():
    'The script entry point'
    ARCTLshell().cmdloop()

if __name__ == "__main__":
    main()



import cmd

from pynusmv.mc import eval_simple_expression
from pynusmv.exception import PyNuSMVError
from ..util.nonExitingArgumentParser import (NonExitingArgumentParser,
                                             ArgumentParsingError)
from pynusmv.exception import NuSMVBddPickingError

def choose_one_state(fsm, BDD, bound=10):
    """
    Use an interactive prompt to ask the user to choose a State in BDD.
    
    fsm -- the model;
    BDD -- a BDD representing states of fsm.
    
    Return a new state of BDD or None if no state has been chosen.
    """
    shell = _One_State_Shell(fsm, BDD, bound)
    shell.cmdloop()
    return shell.chosen
    
    
def choose_next_state(fsm, state, bound=10):
    """
    Use an interactive prompt to ask the user
    to choose a successor of state in fsm.
    
    fsm -- the model;
    state -- a state of fsm.
    
    Return a new (inputs, next) pair of fsm such that next is a successor of
    state in fsm through inputs, or (None, None) if no state has been chosen.
    """
    shell = _Next_State_Shell(fsm, state, bound)
    shell.cmdloop()
    return (shell.chosenInputs, shell.chosen)
    
    
class _One_State_Shell(cmd.Cmd):
    """
    A shell allowing to choose state among a given BDD.
    
    This shell has
        the original BDD (to be able to reset)
        the FSM from which the BDD comes
            (to be able to display and pick states)
        a list of (more and more small) BDDs (to be able to backtrack)
        a bound on the number of displayed states 
            (to display only small amounts of states)
            
    This shell starts with displaying
        either the number of states if the bound is reached
        or all the possible states
            
    This shell provides some commands
        reset (remove all the constraints and restart with original BDD)
        backtrack (remove last constraints from the set of constraints)
        constraint (add constraints)
        choose (choose a state in the ones displayed, if some displayed)
        EOF (quit without getting a state)
        an empty line shows again the last bdd
        
    This shell stops when a state is chosen or EOF is entered.
    When this shell stops, self.chosen is filled with the chosen state, or
    with None if no state has been chosen (EOF entered).
    """
    
    def __init__(self, fsm, bdd, bound=10, prompt=">> "):
        """
        Initialize a one state shell.
        
        fsm -- the FSM of interest;
        bdd -- the original BDD, represents a set of states of fsm;
        bound -- the maximum number of states that can be displayed;
        prompt -- the prompt of the shell.
        """
        super(_One_State_Shell, self).__init__()
        self.prompt = prompt
        self.fsm = fsm
        self.original = bdd
        self.bound = bound
        
        self.constraints = []
        self.bdds = [self.original]
        
        self.shown = [] # The list of last shown states
        self.chosen = None # The chosen state
        
        self.parsers = {} # Parsers for commands
        
        
    def _show_last(self):
        """
        Show the last BDD.
        
        If the size of the last BDD is greater than self.bound,
        show only a message asking for constraints
        (and the size of the BDD in terms of states).
        """
        # Reset shown states
        self.shown = []
        bdd = self.bdds[-1]
        states = self.fsm.pick_all_states(bdd)
        if len(states) > self.bound:
            print("Too many states ("+ str(len(states)) + "), add constraints")
        else:
            prev = None
            for state in states:
                self.shown.append(state)
                # Show the state
                header = " State " + str(len(self.shown)) + " "
                print(header.center(40, "-"))
                values = state.get_str_values()
                for var in values:
                    if (prev is not None and
                        prev.get_str_values()[var] != values[var] or
                        prev is None) :
                        print(var, "=", values[var])
                        
                prev = state
        
        
    def preloop(self):
        self._show_last()
        
    
    # For every command cmd, implement
    #   parse_cmd: create and store the parser for cmd
    #   help_cmd: display the parser help for cmd
    #   do_cmd: actually perform cmd
    
    
    def parse_choose(self):
        """Build and store the parser for the choose command."""
        # choose STATE
        # STATE is an int, corresponding to the chosen state in the previously
        # shown list
        
        if "choose" not in self.parsers:
            parser = NonExitingArgumentParser(
                        "choose",
                        description="Choose a state.",
                        add_help=False)
            parser.add_argument('state', metavar='STATE', type=int,
                                help="the id of the chosen state")
            self.parsers["choose"] = parser
        
        
    def help_choose(self):
        if "choose" not in self.parsers:
            self.parse_choose()
        self.parsers["choose"].print_help()
       
        
    def do_choose(self, arg):
        if "choose" not in self.parsers:
            self.parse_choose()
            
        # Handle arguments parsing errors
        try:
            args = self.parsers["choose"].parse_args(arg.split())
        except ArgumentParsingError as err:
            print(err, end="")
            return False
            
        if len(self.shown) <= 0:
            print("choose: error: no specified states.")
            return False
            
        state = int(args.state) - 1
        if state < 0 or len(self.shown) <= state:
            print("choose: error: unknown state.")
            return False
        else:
            return self._choose(state)
            
    
    def _choose(self, index):
        """
        Choose the state with ID index.
        
        If index is a correct ID, set self.chosen to the state, return True;
        otherwise don't set self.chosen and return False.
        """
        if index < 0 or len(self.shown) <= index:
            print("choose: error: unknown state.")
            return False
        else:
            self.chosen = self.shown[index]
            return True        
            
            
    def emptyline(self):
        """When entering an empty line, the last BDD is shown."""
        self._show_last()
        
        
    def do_backtrack(self, arg):
        """Undo last applied constraints."""
        if len(self.bdds) <= 1:
            print("backtrack: error: cannot backtrack empty constraints.")
        else:
            self.bdds = self.bdds[:-1]
            self.constraints = self.constraints[:-1]
            self._show_last()
        return False
        
        
    def parse_constraint(self):
        """Build and store the parser for the constraints command."""
        # constraints CONSTRAINTS
        # CONSTRAINTS a simple expression representing constraints
        # on the current set of states, used to restrict it
        
        if "constraint" not in self.parsers:
            parser = NonExitingArgumentParser(
                        "constraint",
                        description="Restrict the current BDD"
                                    " with constraints.",
                        add_help=False)
            parser.add_argument('constraints', metavar='CONSTRAINTS', nargs='?',
                                help="the constraints to add; if ommitted, show current list of applied constraints")
            self.parsers["constraint"] = parser
        
    def help_constraint(self):
        if "constraint" not in self.parsers:
            self.parse_constraint()
        self.parsers["constraint"].print_help()
        
    def do_constraint(self, arg):
        if "constraint" not in self.parsers:
            self.parse_constraint()
            
        # Handle arguments parsing errors
        try:
            if arg.strip() == "":
                arg = []
            else:
                arg = [arg.strip()]
            args = self.parsers["constraint"].parse_args(arg)
        except ArgumentParsingError as err:
            print(err, end="")
            return False
            
        if args.constraints is None:
            if len(self.constraints) <= 0:
                print("Currently no constraint.")
            else:
                print("Current constraints:")
                for const in self.constraints:
                    print(const)
            return False
            
        else:
            try:
                # Parse constraints
                constBDD = eval_simple_expression(self.fsm, args.constraints)
                # Restrict the BDD
                newBdd = self.bdds[-1] & constBDD
                # If the BDD is not empty, add it to the list and show it
                if self.fsm.count_states(newBdd) > 0:
                    self.bdds.append(newBdd)
                    self.constraints.append(args.constraints)
                    self._show_last()
                else:
                    # else backtrack (in fact, do nothing)
                    print("constraint: the constraints are too strong,",
                          "no more states; retry.")
            except PyNuSMVError as err:
                print(err)
                return False
            
        return False
        
        
    def do_reset(self, arg):
        """Reset the choosing process."""
        self.constraints = []
        self.bdds = [self.original]
        self._show_last()


    def do_EOF(self, arg):
        "Quit without choosing a state."
        return True
        
        
class _Next_State_Shell(_One_State_Shell):
    """
    A shell allowing to choose a successor of a given state in a given FSM.
    
    This shell differs from _One_State_Shell by
        - getting a state of the FSM instead of a BDD;
        - choosing a successor of this state;
        - displaying possible inputs to reach each possible state;
        - chosen item can be a tuple (inputs, state) instead of just a state,
          if there are input variables in the model. In this case, self.chosen
          is the state and self.chosenInputs is the inputs (can be None).
    """
    
    def __init__(self, fsm, state, bound=10, prompt=">> "):
        """
        Initialize a one state shell.
        
        fsm -- the FSM of interest;
        bdd -- the state of interest, belongs to fsm;
        bound -- the maximum number of states that can be displayed;
        prompt -- the prompt of the shell.
        """
        
        self.state = state
        self.chosenInputs = None
        bdd = fsm.post(state)
        super(_Next_State_Shell, self).__init__(fsm, bdd, bound, prompt)
        
        
    def _show_last(self):
        """
        Show the last BDD.
        
        If the size of the last BDD is greater than self.bound,
        show only a message asking for constraints
        (and the size of the BDD in terms of states).
        """
        
        def show_state_or_inputs(item, prev = None):
            values = item.get_str_values()
            for var in values:
                if (prev is not None and
                    prev.get_str_values()[var] != values[var] or
                    prev is None) :
                    print(var, "=", values[var])
        
        # Reset shown states
        self.shown = []
        bdd = self.bdds[-1]
        states = self.fsm.pick_all_states(bdd)
        if len(states) > self.bound:
            print("Too many states ("+ str(len(states)) + "), add constraints")
        else:
            prev = None
            for state in states:
                # Get inputs
                inputs = self.fsm.get_inputs_between_states(self.state, state)
                
                try:
                    inputs = self.fsm.pick_all_inputs(inputs)
                    
                    print(" State ".center(40, "-"))
                    show_state_or_inputs(state, prev)
                    print(" Reachable through ".center(40,"-"))
                    previ = None
                    for inp in inputs:
                        self.shown.append((inp,state))
                        print((" Inputs " + str(len(self.shown)) + " ").
                              center(40, "-"))
                        show_state_or_inputs(inp, previ)
                        previ = inp
                    
                except NuSMVBddPickingError:
                    # Cannot get inputs, so no inputs
                    self.shown.append((None, state))
                    # Show the state
                    header = " State " + str(len(self.shown)) + " "
                    print(header.center(40, "-"))
                    show_state_or_inputs(state, prev)
                        
                prev = state
                
                
    def _choose(self, index):
        """
        Choose the (inputs, state) pair with ID index.
        
        If index is a correct ID, set self.chosen to the state and self.inputs
        to the inputs, return True;
        otherwise don't set self.chosen and self.inputs and return False.
        """
        if index < 0 or len(self.shown) <= index:
            print("choose: error: unknown state.")
            return False
        else:
            self.chosenInputs, self.chosen = self.shown[index]
            return True


from ..ast import (TrueExp, FalseExp, Init, Reachable,
                   Atom, Not, And, Or, Implies, Iff, 
                   AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                   nK, nE, nD, nC, K, E, D, C)
from ..eval import evalCTLK
from ..explain import (explain_ex, explain_eg, explain_eu,
                       explain_nk, explain_ne, explain_nd, explain_nc,
                       explain_reachable)
from .tlace import Tlacenode, TemporalBranch, EpistemicBranch

def explain_witness(fsm, state, spec):
    """
    Explain why state of fsm satisfies spec.
    
    state must satisfy spec in fsm. No check is made to ensure that.
    Return a TLACE node explaining why state of fsm satifies spec.
    """
    
    if type(spec) is TrueExp:
        # state is its own explanation, no need for annotation
        return Tlacenode(state, None, None, None)
        
    elif type(spec) is FalseExp:    
        print("[ERROR] CTLK TLACE explain_witness:",
              "cannot explain why state satisfies False",
              spec)
        # TODO Error, cannot explain why state |= False
        return None
        
    elif type(spec) is Init:
        return Tlacenode(state, (spec,), None, None)
    
    elif type(spec) is Atom:
        # state is its own explanation
        return Tlacenode(state, (spec,), None, None)
        
    elif type(spec) is Not:
        return explain_countex(fsm, state, spec.child)
        
    elif type(spec) is And:
        # Get left and right explanations, then merge them
        left = explain_witness(fsm, state, spec.left)
        right = explain_witness(fsm, state, spec.right)
        return Tlacenode(state,
                         left.atomics + right.atomics,
                         left.branches + right.branches,
                         left.universals + right.universals
                        )
        
    elif type(spec) is Or:
        # If state satisfies spec.left, explain it
        # otherwise, state satisfies spec.right, so explain it
        specbdd = evalCTLK(fsm, spec.left)
        if state <= specbdd:
            return explain_witness(fsm, state, spec.left)
        else:
            return explain_witness(fsm, state, spec.right)
        
    elif type(spec) is Implies:
        # a -> b is ~a | b
        return explain_witness(fsm, state, Or(Not(spec.left), spec.right))
        
    elif type(spec) is Iff:
        # a <-> b is (a & b) | (~a & ~b)
        return explain_witness(
                fsm, state,
                Or(And(spec.left, spec.right),
                   And(Not(spec.left), Not(spec.right)))
               )
        
    elif type(spec) in {AF, AG, AX, AU, AW, K, E, D, C}:
        # Cannot explain with a single path
        return Tlacenode(state, None, None, (spec,))
                       
    elif type(spec) in {EF, EG, EX, EU, EW, nK, nE, nD, nC, Reachable}:
        branch = explain_branch(fsm, state, spec, spec)
        return Tlacenode(state, None, (branch,), None)
        
    else:
        # TODO Generate error
        print("[ERROR] CTLK TLACE explain_witness:",
              "unrecognized specification type",
              spec)
        return None
        
        
def explain_countex(fsm, state, spec):
    """
    Explain why state of fsm violates spec.
    
    state must not satisfy spec in fsm. No check is made to ensure that.
    Return a TLACE node explaining why state of fsm violates spec.
    """
    
    if type(spec) is TrueExp:    
        print("[ERROR] CTLK TLACE explain_countex:",
              "cannot explain why state violates True",
              spec)
        # TODO Error, cannot explain why state |/= True
        return None
        
    elif type(spec) is FalseExp:
        return explain_witness(fsm, state, TrueExp())
        
    elif type(spec) is Init:
        return Tlacenode(state, (Not(spec),), None, None)
        
    elif type(spec) is Atom:
        # state is its own explanation
        return Tlacenode(state, (Not(spec),), None, None)
        
    elif type(spec) is Not:
        return explain_witness(fsm, state, spec.child)
        
    elif type(spec) is And:
        # ~(a & b) = ~a | ~b
        return explain_witness(fsm, state, Or(Not(spec.left), Not(spec.right)))
        
    elif type(spec) is Or:
        # ~(a | b) = ~a & ~b
        return explain_witness(fsm, state, And(Not(spec.left), Not(spec.right)))
        
    elif type(spec) is Implies:
        # ~(a -> b) = a & ~b
        return explain_witness(fsm, state, And(spec.left, Not(spec.right)))
        
    elif type(spec) is Iff:
        # ~(a <-> b) = (a & ~b) | (~a & b)
        return explain_witness(fsm, state,
                               Or(And(spec.left, Not(spec.right)),
                                  And(Not(spec.left), spec.right)))
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        return explain_witness(fsm, state, EG(Not(spec.child)))
        
    elif type(spec) is AG:
        # AG p = ~EF ~p
        return explain_witness(fsm, state, EF(Not(spec.child)))
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        return explain_witness(fsm, state, EX(Not(spec.child)))
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q]
        return explain_witness(fsm, state,
                               EW(Not(spec.right),
                                  And(Not(spec.left), Not(spec.right))))
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        return explain_witness(fsm, state,
                               EU(Not(spec.right),
                                  And(Not(spec.left), Not(spec.right))))
                     
    elif type(spec) is EF:
        # EF p = ~AG ~p
        return explain_witness(fsm, state, AG(Not(spec.child)))
        
    elif type(spec) is EG:
        # EG p = ~AF ~p
        return explain_witness(fsm, state, AF(Not(spec.child)))
        
    elif type(spec) is EX:
        # EX p = ~AX ~p
        return explain_witness(fsm, state, AX(Not(spec.child)))
        
    elif type(spec) is EU:
        # E[p U q] = ~A[~q W ~p & ~q]
        return explain_witness(fsm, state,
                               AW(Not(spec.right),
                                  And(Not(spec.left), Not(spec.right))))
        
    elif type(spec) is EW:
        # E[p W q] = ~A[~q U ~p & ~q]
        return explain_witness(fsm, state,
                               AU(Not(spec.right),
                                  And(Not(spec.left), Not(spec.right))))
                                  
    elif type(spec) is nK:
        # nK<ag> p = ~K<ag> ~p
        return explain_witness(fsm, state, K(spec.agent, Not(spec.child)))
                                  
    elif type(spec) is nE:
        # nE<group> p = ~E<group> ~p
        return explain_witness(fsm, state, E(spec.group, Not(spec.child)))
                                  
    elif type(spec) is nD:
        # nD<group> p = ~D<group> ~p
        return explain_witness(fsm, state, D(spec.group, Not(spec.child)))
                                  
    elif type(spec) is nC:
        # nC<group> p = ~C<group> ~p
        return explain_witness(fsm, state, C(spec.group, Not(spec.child)))
                                  
    elif type(spec) is K:
        # K<ag> p = ~nK<ag> ~p
        return explain_witness(fsm, state, nK(spec.agent, Not(spec.child)))
                                  
    elif type(spec) is E:
        # E<group> p = ~nE<group> ~p
        return explain_witness(fsm, state, nE(spec.group, Not(spec.child)))
                                  
    elif type(spec) is D:
        # D<group> p = ~nD<group> ~p
        return explain_witness(fsm, state, nD(spec.group, Not(spec.child)))
                                  
    elif type(spec) is C:
        # C<group> p = ~nC<group> ~p
        return explain_witness(fsm, state, nC(spec.group, Not(spec.child)))
        
    elif type(spec) is Reachable:
        # Is its own explanation
        return Tlacenode(state, None, None, (Not(spec),))
        
    else:
        # TODO Generate error
        print("[ERROR] CTLK explain_countex: unrecognized specification type",
              spec)
        return None
        
        
def explain_branch(fsm, state, spec, originalspec):
    """
    Return a TLACE branch explaining why state of fsm satisfies spec.
    
    state must satisfy spec in fsm. No check is made to ensure that.
    spec must be an existential operator (EX, EF, EU, EG, EW, nK, nE, nD, nC).
    originalspec is the specification with which the created branch will be
    annotated.
    """
    if type(spec) is EX:
        phi = evalCTLK(fsm, spec.child)
        path = explain_ex(fsm, state, phi)
        branch = (Tlacenode(path[0]),
                  path[1],
                  explain_witness(fsm, path[2], spec.child))
        return TemporalBranch(originalspec, branch)

    elif type(spec) is EF:
        return explain_branch(fsm, state,
                              EU(TrueExp(), spec.child),
                              originalspec)

    elif type(spec) is EG:
        phi = evalCTLK(fsm, spec.child)
        (path, (inloop, loopstate)) = explain_eg(fsm, state, phi)

        branch = []
        # intermediate states
        loop = None
        for s, i in zip(path[::2], path[1::2]):
            wit = explain_witness(fsm, s, spec.child)
            branch.append(wit)
            branch.append(i)
            # manage the loop
            if loopstate == s:
                loop = wit
                
        # last state and loop
        wit = explain_witness(fsm, path[-1], spec.child)
        branch.append(wit)
        if loopstate == path[-1]:
            loop = wit
        finalloop = (inloop, loop)
        
        return TemporalBranch(originalspec, tuple(branch), finalloop)

    elif type(spec) is EU:
        phi = evalCTLK(fsm, spec.left)
        psi = evalCTLK(fsm, spec.right)
        path = explain_eu(fsm, state, phi, psi)

        branch = []
        # intermediate states
        for s, i in zip(path[::2], path[1::2]):
            branch.append(explain_witness(fsm, s, spec.left))
            branch.append(i)
        # last state
        branch.append(explain_witness(fsm, path[-1], spec.right))

        return TemporalBranch(originalspec, tuple(branch))

    elif type(spec) is EW:
        # E[p W q] = E[p U q] | EG p
        euspec = EU(spec.left, spec.right)
        egspec = EG(spec.left)
        if state <= evalCTLK(fsm, euspec):
            return explain_branch(fsm, state, euspec, originalspec)
        else:
            return explain_branch(fsm, state, egspec, originalspec)
            
    elif type(spec) is Reachable:
        # Get the inversed path
        path = explain_reachable(fsm, state)
        
        # Construct the path of nodes
        branch = []
        for (s, i) in zip(path[::2], path[1::2]):
            branch.append(Tlacenode(s))
            branch.append(i)
        # Special case for the last node: it is Init
        branch.append(explain_witness(fsm, path[-1], Init()))
        
        return TemporalBranch(originalspec, tuple(branch))
            
    elif type(spec) is nK:        
        # Get the equivalent state
        phi = evalCTLK(fsm, spec.child)
        path = explain_nk(fsm, state, spec.agent.value, phi)
        
        # Explain why the equiv state satisfies Reachable and phi
        # and construct the branch
        branch = [Tlacenode(path[0]), path[1],
                  explain_witness(fsm, path[2], And(Reachable(), spec.child))]
        
        # Return the epistemic branch
        return EpistemicBranch(originalspec, tuple(branch))
        
    elif type(spec) is nE:        
        # Get the equivalent state
        phi = evalCTLK(fsm, spec.child)
        path = explain_ne(fsm, state, [ag.value for ag in spec.group], phi)
        
        # Explain why the equiv state satisfies Reachable and phi
        # and construct the branch
        branch = [Tlacenode(path[0]), path[1],
                  explain_witness(fsm, path[2], And(Reachable(), spec.child))]
        
        # Return the epistemic branch
        return EpistemicBranch(originalspec, tuple(branch))
        
    elif type(spec) is nD:        
        # Get the equivalent state
        phi = evalCTLK(fsm, spec.child)
        path = explain_nd(fsm, state, [ag.value for ag in spec.group], phi)
        
        # Explain why the equiv state satisfies Reachable and phi
        # and construct the branch
        branch = [Tlacenode(path[0]), path[1],
                  explain_witness(fsm, path[2], And(Reachable(), spec.child))]
        
        # Return the epistemic branch
        return EpistemicBranch(originalspec, tuple(branch))
        
    if type(spec) is nC:
        # Get the knowledge path
        phi = evalCTLK(fsm, spec.child)
        path = explain_nc(fsm, state, [ag.value for ag in spec.group], phi)
        
        # Discard the first element of path: this is state and does not have
        # to be explained reachable
        branch = [Tlacenode(path[0]), path[1]]
        path = path[2:]
        
        # Build the branch : show that all intermediate states are reachable
        for (s, ag) in zip(path[::2], path[1::2]):
            branch.append(explain_witness(fsm, s, Reachable()))
            branch.append(ag)
        
        # Complete the branch : show that the last one is reachable
        # satisfies phi
        branch.append(explain_witness(fsm, path[-1],
                                                  And(Reachable(), spec.child)))
        
        # Return the epistemic branch
        return EpistemicBranch(originalspec, tuple(branch))

    else:
        # TODO Generate error
        print("[ERROR] CTLK explain_branch:",
              "unrecognized specification type",
              spec)
        return None


'''
This module tests the behavior of the functionalities defined in the parsing
module :mod:`pynusmv_tools.bmcLTL.parsing`
'''
from unittest              import TestCase

from pynusmv_tools.bmcLTL          import parsing 

class TestParsing(TestCase):
    
    def test_parse_variable(self):
        ast = parsing.parseLTL("phi")
        self.assertEqual("Proposition(phi)", str(ast))
        
    def test_parse_variable_dotted_name(self):
        ast = parsing.parseLTL("p1.waiting")
        self.assertEqual("Proposition(p1.waiting)", str(ast))
    
    def test_parse_variable_arobase_name(self):
        ast = parsing.parseLTL("p1@waiting")
        self.assertEqual("Proposition(p1@waiting)", str(ast))
        
    def test_parse_variable_specialchar_name(self):
        ast = parsing.parseLTL("p1~waiting")
        # everything from the unknown punctuation on is discarded
        self.assertEqual("Proposition(p1)", str(ast))
    
    def test_parse_variable_containing_TRUE(self):
        ast = parsing.parseLTL("@TRUE")
        # everything from the unknown punctuation on is discarded
        self.assertEqual("Proposition(@TRUE)", str(ast))
        
        ast = parsing.parseLTL("TRUE@")
        # everything from the unknown punctuation on is discarded
        self.assertEqual("Proposition(TRUE@)", str(ast))
        
        ast = parsing.parseLTL("TRUE.")
        # everything from the unknown punctuation on is discarded
        self.assertEqual("Proposition(TRUE.)", str(ast))
        
    def test_parse_variable_containing_FALSE(self):
        ast = parsing.parseLTL("@FALSE")
        # everything from the unknown punctuation on is discarded
        self.assertEqual("Proposition(@FALSE)", str(ast))
        
        ast = parsing.parseLTL("FALSE@")
        # everything from the unknown punctuation on is discarded
        self.assertEqual("Proposition(FALSE@)", str(ast))
        
        ast = parsing.parseLTL("FALSE.")
        # everything from the unknown punctuation on is discarded
        self.assertEqual("Proposition(FALSE.)", str(ast))
    
    def test_parse_constant(self):
        ast = parsing.parseLTL("TRUE")
        self.assertEqual("Constant(TRUE)", str(ast))
        #
        ast = parsing.parseLTL("FALSE")
        self.assertEqual("Constant(FALSE)", str(ast))
        
    def test_parse_number(self):
        # question, shouldn't we reject this instead ?
        ast = parsing.parseLTL("1256")
        self.assertEqual("Constant(1256)", str(ast))
        
    def test_arithmetic_unary_minus(self):
        # number can be negated
        ast = parsing.parseLTL("-12")
        self.assertEqual("Proposition(- (12))", str(ast))
        # and so does a variable
        ast = parsing.parseLTL("-a")
        self.assertEqual("Proposition(- (a))", str(ast))
        # or a nested expression
        ast = parsing.parseLTL("-( 6 * 8)")
        self.assertEqual("Proposition(- ((6) * (8)))", str(ast))
    
    def test_arithmetic_multiplication(self):    
        # numbers 
        ast = parsing.parseLTL("6 * 8")
        self.assertEqual("Proposition((6) * (8))", str(ast))
        # vars
        ast = parsing.parseLTL("a * b")
        self.assertEqual("Proposition((a) * (b))", str(ast))
        # chainable
        ast = parsing.parseLTL("a * b * c")
        self.assertEqual("Proposition((a) * (b) * (c))", str(ast))
        # priority over +/-
        ast = parsing.parseLTL("a + b * c")
        self.assertEqual("Proposition((a) + ((b) * (c)))", str(ast))
        ast = parsing.parseLTL("a * b - c")
        self.assertEqual("Proposition(((a) * (b)) - (c))", str(ast))
    
    def test_arithmetic_division(self):    
        # numbers 
        ast = parsing.parseLTL("6 / 8")
        self.assertEqual("Proposition((6) / (8))", str(ast))
        # vars
        ast = parsing.parseLTL("a / b")
        self.assertEqual("Proposition((a) / (b))", str(ast))
        # chainable
        ast = parsing.parseLTL("a / b / c")
        self.assertEqual("Proposition((a) / (b) / (c))", str(ast))
        # priority over +/-
        ast = parsing.parseLTL("a + b / c")
        self.assertEqual("Proposition((a) + ((b) / (c)))", str(ast))
        ast = parsing.parseLTL("a / b - c")
        self.assertEqual("Proposition(((a) / (b)) - (c))", str(ast))
        
    def test_arithmetic_modulus(self):    
        # numbers 
        ast = parsing.parseLTL("6 mod 8")
        self.assertEqual("Proposition((6) mod (8))", str(ast))
        # vars
        ast = parsing.parseLTL("a mod b")
        self.assertEqual("Proposition((a) mod (b))", str(ast))
        # chainable
        ast = parsing.parseLTL("a mod b mod c")
        self.assertEqual("Proposition((a) mod (b) mod (c))", str(ast))
        # priority over +/-
        ast = parsing.parseLTL("a + b mod c")
        self.assertEqual("Proposition((a) + ((b) mod (c)))", str(ast))
        ast = parsing.parseLTL("a mod b - c")
        self.assertEqual("Proposition(((a) mod (b)) - (c))", str(ast))
        
    def test_arithmetic_multiplicatives_can_be_mixed_at_will(self):
        ast = parsing.parseLTL("a * b / c mod d * b mod a / c")
        self.assertEqual("Proposition((a) * (b) / (c) mod (d) * (b) mod (a) / (c))", str(ast))
        
    def test_arithmetic_addition(self):
        # numbers 
        ast = parsing.parseLTL("6 + 8")
        self.assertEqual("Proposition((6) + (8))", str(ast))
        
        ast = parsing.parseLTL("-6 + 8")
        self.assertEqual("Proposition((- (6)) + (8))", str(ast))
        
        # vars
        ast = parsing.parseLTL("a + b")
        self.assertEqual("Proposition((a) + (b))", str(ast))
        # chainable
        ast = parsing.parseLTL("a + b + c")
        self.assertEqual("Proposition((a) + (b) + (c))", str(ast))
        # priority over <</>>
        ast = parsing.parseLTL("a + b << c")
        self.assertEqual("Proposition(((a) + (b)) << (c))", str(ast))
        ast = parsing.parseLTL("a + b >> c")
        self.assertEqual("Proposition(((a) + (b)) >> (c))", str(ast))
        
    def test_arithmetic_subtraction(self):
        # numbers 
        ast = parsing.parseLTL("6 - 8")
        self.assertEqual("Proposition((6) - (8))", str(ast))
        # vars
        ast = parsing.parseLTL("a - b")
        self.assertEqual("Proposition((a) - (b))", str(ast))
        # chainable
        ast = parsing.parseLTL("a - b - c")
        self.assertEqual("Proposition((a) - (b) - (c))", str(ast))
        # priority over <</>>
        ast = parsing.parseLTL("a - b << c")
        self.assertEqual("Proposition(((a) - (b)) << (c))", str(ast))
        ast = parsing.parseLTL("a - b >> c")
        self.assertEqual("Proposition(((a) - (b)) >> (c))", str(ast))
    
    def test_arithmetic_additives_can_be_mixed_at_will(self):
        ast = parsing.parseLTL("a + b - c - d + b + a - c")
        self.assertEqual("Proposition((a) + (b) - (c) - (d) + (b) + (a) - (c))", str(ast))
        
    def test_arithmetic_lshift(self):
        # numbers 
        ast = parsing.parseLTL("6 << 8")
        self.assertEqual("Proposition((6) << (8))", str(ast))
        # vars
        ast = parsing.parseLTL("a << b")
        self.assertEqual("Proposition((a) << (b))", str(ast))
        # chainable
        ast = parsing.parseLTL("a << b << c")
        self.assertEqual("Proposition((a) << (b) << (c))", str(ast))
        # priority over comparison
        ast = parsing.parseLTL("a << b <= c")
        self.assertEqual("Proposition(((a) << (b)) <= (c))", str(ast))
        ast = parsing.parseLTL("a << b = c")
        self.assertEqual("Proposition(((a) << (b)) = (c))", str(ast))
        
    def test_arithmetic_rshift(self):
        # numbers 
        ast = parsing.parseLTL("6 >> 8")
        self.assertEqual("Proposition((6) >> (8))", str(ast))
        # vars
        ast = parsing.parseLTL("a >> b")
        self.assertEqual("Proposition((a) >> (b))", str(ast))
        # chainable
        ast = parsing.parseLTL("a >> b >> c")
        self.assertEqual("Proposition((a) >> (b) >> (c))", str(ast))
        
        # priority over comparison
        ast = parsing.parseLTL("a >> b <= c")
        self.assertEqual("Proposition(((a) >> (b)) <= (c))", str(ast))
        ast = parsing.parseLTL("a >> b = c")
        self.assertEqual("Proposition(((a) >> (b)) = (c))", str(ast))
        
    def test_shifts_can_be_mixed_at_will(self):
        ast = parsing.parseLTL("a << b << c >> d << b >> a >> c")
        self.assertEqual("Proposition((a) << (b) << (c) >> (d) << (b) >> (a) >> (c))", str(ast))
    
    def test_comparison(self):
        # numbers 
        ast = parsing.parseLTL("6 < 8")
        self.assertEqual("Proposition((6) < (8))", str(ast))
        
        ast = parsing.parseLTL("6 <= 8")
        self.assertEqual("Proposition((6) <= (8))", str(ast))
        
        ast = parsing.parseLTL("6 > 8")
        self.assertEqual("Proposition((6) > (8))", str(ast))
        
        ast = parsing.parseLTL("6 >= 8")
        self.assertEqual("Proposition((6) >= (8))", str(ast))
        
        ast = parsing.parseLTL("6 = 8")
        self.assertEqual("Proposition((6) = (8))", str(ast))
        
        ast = parsing.parseLTL("6 != 8")
        self.assertEqual("Proposition((6) != (8))", str(ast))
        
        # vars
        ast = parsing.parseLTL("a < b")
        self.assertEqual("Proposition((a) < (b))", str(ast))
        
        ast = parsing.parseLTL("a <= b")
        self.assertEqual("Proposition((a) <= (b))", str(ast))
        
        ast = parsing.parseLTL("a > b")
        self.assertEqual("Proposition((a) > (b))", str(ast))
        
        ast = parsing.parseLTL("a >= b")
        self.assertEqual("Proposition((a) >= (b))", str(ast))
        
        ast = parsing.parseLTL("a = b")
        self.assertEqual("Proposition((a) = (b))", str(ast))
        
        ast = parsing.parseLTL("a != b")
        self.assertEqual("Proposition((a) != (b))", str(ast))
        
        # chainable
        ast = parsing.parseLTL("a < b < c")
        self.assertEqual("Proposition((a) < (b) < (c))", str(ast))
        
        ast = parsing.parseLTL("a <= b <= c")
        self.assertEqual("Proposition((a) <= (b) <= (c))", str(ast))
        
        ast = parsing.parseLTL("a > b > c")
        self.assertEqual("Proposition((a) > (b) > (c))", str(ast))
        
        ast = parsing.parseLTL("a >= b >= c")
        self.assertEqual("Proposition((a) >= (b) >= (c))", str(ast))
        
        ast = parsing.parseLTL("a = b = c")
        self.assertEqual("Proposition((a) = (b) = (c))", str(ast))
        
        ast = parsing.parseLTL("a != b != c")
        self.assertEqual("Proposition((a) != (b) != (c))", str(ast))
        
        # mixable
        ast = parsing.parseLTL("a != b < c >= d")
        self.assertEqual("Proposition((a) != (b) < (c) >= (d))", str(ast))
    
    def test_parse_not(self):
        ast = parsing.parseLTL("! phi")
        self.assertEqual("(Not Proposition(phi))", str(ast))
        
        # spacing is irrelevant
        ast = parsing.parseLTL("!phi")
        self.assertEqual("(Not Proposition(phi))", str(ast))
        
        # can be chained
        ast = parsing.parseLTL("! ! ! phi")
        self.assertEqual("(Not (Not (Not Proposition(phi))))", str(ast))
        
        # space is irrelevant (even when chained)
        ast = parsing.parseLTL("!!!phi")
        self.assertEqual("(Not (Not (Not Proposition(phi))))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("!TRUE")
        self.assertEqual("(Not Constant(TRUE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("!(a | b)")
        self.assertEqual("(Not (Proposition(a) Or Proposition(b)))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("![]a")
        self.assertEqual("(Not (Globally Proposition(a)))", str(ast))

    def test_parse_and(self):
        ast = parsing.parseLTL("phi & psi")
        self.assertEqual("(Proposition(phi) And Proposition(psi))", str(ast))
        
        # spacing is irrelevant
        ast = parsing.parseLTL("phi&psi")
        self.assertEqual("(Proposition(phi) And Proposition(psi))", str(ast))
        
        # can be chained (right associative)
        ast = parsing.parseLTL("phi&psi&chi")
        self.assertEqual("((Proposition(phi) And Proposition(psi)) And Proposition(chi))", str(ast))
        
        # space is irrelevant (even when chained)
        ast = parsing.parseLTL("phi & psi & chi")
        self.assertEqual("((Proposition(phi) And Proposition(psi)) And Proposition(chi))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("TRUE & FALSE")
        self.assertEqual("(Constant(TRUE) And Constant(FALSE))", str(ast))
        ast = parsing.parseLTL("TRUE & alpha")
        self.assertEqual("(Constant(TRUE) And Proposition(alpha))", str(ast))
        ast = parsing.parseLTL("alpha & FALSE")
        self.assertEqual("(Proposition(alpha) And Constant(FALSE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("(a | b) & x")
        self.assertEqual("((Proposition(a) Or Proposition(b)) And Proposition(x))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("[]a & <>b")
        self.assertEqual("((Globally Proposition(a)) And (Eventually Proposition(b)))", str(ast))
        ast = parsing.parseLTL("[]a & b")
        self.assertEqual("((Globally Proposition(a)) And Proposition(b))", str(ast))
        ast = parsing.parseLTL("a & <>b")
        self.assertEqual("(Proposition(a) And (Eventually Proposition(b)))", str(ast))
        
    def test_parse_or(self):
        ast = parsing.parseLTL("phi | psi")
        self.assertEqual("(Proposition(phi) Or Proposition(psi))", str(ast))
        # spacing is irrelevant
        ast = parsing.parseLTL("phi|psi")
        self.assertEqual("(Proposition(phi) Or Proposition(psi))", str(ast))
        # can be chained (left associative)
        ast = parsing.parseLTL("phi | psi | chi")
        self.assertEqual("((Proposition(phi) Or Proposition(psi)) Or Proposition(chi))", str(ast))
        # space is irrelevant (even when chained)
        ast = parsing.parseLTL("phi|psi|chi")
        self.assertEqual("((Proposition(phi) Or Proposition(psi)) Or Proposition(chi))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("TRUE | FALSE")
        self.assertEqual("(Constant(TRUE) Or Constant(FALSE))", str(ast))
        ast = parsing.parseLTL("TRUE | alpha")
        self.assertEqual("(Constant(TRUE) Or Proposition(alpha))", str(ast))
        ast = parsing.parseLTL("alpha | FALSE")
        self.assertEqual("(Proposition(alpha) Or Constant(FALSE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("(a | b) | x")
        self.assertEqual("((Proposition(a) Or Proposition(b)) Or Proposition(x))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("[]a | <>b")
        self.assertEqual("((Globally Proposition(a)) Or (Eventually Proposition(b)))", str(ast))
        ast = parsing.parseLTL("[]a | b")
        self.assertEqual("((Globally Proposition(a)) Or Proposition(b))", str(ast))
        ast = parsing.parseLTL("a | <>b")
        self.assertEqual("(Proposition(a) Or (Eventually Proposition(b)))", str(ast))
        
    def test_parse_xor(self):
        ast = parsing.parseLTL("phi ^ psi")
        self.assertEqual("(Proposition(phi) Xor Proposition(psi))", str(ast))
        # spacing is irrelevant
        ast = parsing.parseLTL("phi^psi")
        self.assertEqual("(Proposition(phi) Xor Proposition(psi))", str(ast))
        # can be chained (left associative)
        ast = parsing.parseLTL("phi ^ psi ^ chi")
        self.assertEqual("((Proposition(phi) Xor Proposition(psi)) Xor Proposition(chi))", str(ast))
        # space is irrelevant (even when chained)
        ast = parsing.parseLTL("phi^psi^chi")
        self.assertEqual("((Proposition(phi) Xor Proposition(psi)) Xor Proposition(chi))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("TRUE ^ FALSE")
        self.assertEqual("(Constant(TRUE) Xor Constant(FALSE))", str(ast))
        ast = parsing.parseLTL("TRUE ^ alpha")
        self.assertEqual("(Constant(TRUE) Xor Proposition(alpha))", str(ast))
        ast = parsing.parseLTL("alpha ^ FALSE")
        self.assertEqual("(Proposition(alpha) Xor Constant(FALSE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("(a ^ b) ^ x")
        self.assertEqual("((Proposition(a) Xor Proposition(b)) Xor Proposition(x))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("[]a ^ <>b")
        self.assertEqual("((Globally Proposition(a)) Xor (Eventually Proposition(b)))", str(ast))
        ast = parsing.parseLTL("[]a ^ b")
        self.assertEqual("((Globally Proposition(a)) Xor Proposition(b))", str(ast))
        ast = parsing.parseLTL("a ^ <>b")
        self.assertEqual("(Proposition(a) Xor (Eventually Proposition(b)))", str(ast))
        
    def test_parse_imply(self):
        ast = parsing.parseLTL("phi => psi")
        self.assertEqual("(Proposition(phi) Imply Proposition(psi))", str(ast))
        # spacing is irrelevant
        ast = parsing.parseLTL("phi=>psi")
        self.assertEqual("(Proposition(phi) Imply Proposition(psi))", str(ast))
        # can be chained (right associative)
        ast = parsing.parseLTL("phi => psi => chi")
        self.assertEqual("((Proposition(phi) Imply Proposition(psi)) Imply Proposition(chi))", str(ast))
        # space is irrelevant (even when chained)
        ast = parsing.parseLTL("phi=>psi=>chi")
        self.assertEqual("((Proposition(phi) Imply Proposition(psi)) Imply Proposition(chi))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("TRUE => FALSE")
        self.assertEqual("(Constant(TRUE) Imply Constant(FALSE))", str(ast))
        ast = parsing.parseLTL("TRUE => alpha")
        self.assertEqual("(Constant(TRUE) Imply Proposition(alpha))", str(ast))
        ast = parsing.parseLTL("alpha => FALSE")
        self.assertEqual("(Proposition(alpha) Imply Constant(FALSE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("(a | b) => x")
        self.assertEqual("((Proposition(a) Or Proposition(b)) Imply Proposition(x))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("[]a => <>b")
        self.assertEqual("((Globally Proposition(a)) Imply (Eventually Proposition(b)))", str(ast))
        ast = parsing.parseLTL("[]a => b")
        self.assertEqual("((Globally Proposition(a)) Imply Proposition(b))", str(ast))
        ast = parsing.parseLTL("a => <>b")
        self.assertEqual("(Proposition(a) Imply (Eventually Proposition(b)))", str(ast))
        
    def test_parse_iff(self):
        ast = parsing.parseLTL("phi <=> psi")
        self.assertEqual("(Proposition(phi) Equiv Proposition(psi))", str(ast))
        # spacing is irrelevant
        ast = parsing.parseLTL("phi<=>psi")
        self.assertEqual("(Proposition(phi) Equiv Proposition(psi))", str(ast))
        # can be chained (right associative)
        ast = parsing.parseLTL("phi <=> psi <=> chi")
        self.assertEqual("((Proposition(phi) Equiv Proposition(psi)) Equiv Proposition(chi))", str(ast))
        # space is irrelevant (even when chained)
        ast = parsing.parseLTL("phi<=>psi<=>chi")
        self.assertEqual("((Proposition(phi) Equiv Proposition(psi)) Equiv Proposition(chi))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("TRUE <=> FALSE")
        self.assertEqual("(Constant(TRUE) Equiv Constant(FALSE))", str(ast))
        ast = parsing.parseLTL("TRUE <=> alpha")
        self.assertEqual("(Constant(TRUE) Equiv Proposition(alpha))", str(ast))
        ast = parsing.parseLTL("alpha <=> FALSE")
        self.assertEqual("(Proposition(alpha) Equiv Constant(FALSE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("(a | b) <=> x")
        self.assertEqual("((Proposition(a) Or Proposition(b)) Equiv Proposition(x))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("[]a <=> <>b")
        self.assertEqual("((Globally Proposition(a)) Equiv (Eventually Proposition(b)))", str(ast))
        ast = parsing.parseLTL("[]a <=> b")
        self.assertEqual("((Globally Proposition(a)) Equiv Proposition(b))", str(ast))
        ast = parsing.parseLTL("a <=> <>b")
        self.assertEqual("(Proposition(a) Equiv (Eventually Proposition(b)))", str(ast))
        
    def test_parse_globally(self):
        ast = parsing.parseLTL("[] phi")
        self.assertEqual("(Globally Proposition(phi))", str(ast))
        
        # spacing is irrelevant
        ast = parsing.parseLTL("[]phi")
        self.assertEqual("(Globally Proposition(phi))", str(ast))
        
        # can be chained
        ast = parsing.parseLTL("[] [] [] phi")
        self.assertEqual("(Globally (Globally (Globally Proposition(phi))))", str(ast))
        
        # space is irrelevant (even when chained)
        ast = parsing.parseLTL("[][][]phi")
        self.assertEqual("(Globally (Globally (Globally Proposition(phi))))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("[]TRUE")
        self.assertEqual("(Globally Constant(TRUE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("[](a | b)")
        self.assertEqual("(Globally (Proposition(a) Or Proposition(b)))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("[]<>a")
        self.assertEqual("(Globally (Eventually Proposition(a)))", str(ast))
    
    def test_parse_eventually(self):
        ast = parsing.parseLTL("<> phi")
        self.assertEqual("(Eventually Proposition(phi))", str(ast))
        
        # spacing is irrelevant
        ast = parsing.parseLTL("<>phi")
        self.assertEqual("(Eventually Proposition(phi))", str(ast))
        
        # can be chained
        ast = parsing.parseLTL("<> <> <> phi")
        self.assertEqual("(Eventually (Eventually (Eventually Proposition(phi))))", str(ast))
        
        # space is irrelevant (even when chained)
        ast = parsing.parseLTL("<><><>phi")
        self.assertEqual("(Eventually (Eventually (Eventually Proposition(phi))))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("<>TRUE")
        self.assertEqual("(Eventually Constant(TRUE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("<>(a | b)")
        self.assertEqual("(Eventually (Proposition(a) Or Proposition(b)))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("<>[]a")
        self.assertEqual("(Eventually (Globally Proposition(a)))", str(ast))
        
    def test_parse_next(self):
        ast = parsing.parseLTL("() phi")
        self.assertEqual("(Next Proposition(phi))", str(ast))
        
        # spacing is irrelevant
        ast = parsing.parseLTL("()phi")
        self.assertEqual("(Next Proposition(phi))", str(ast))
        
        # can be chained
        ast = parsing.parseLTL("() () () phi")
        self.assertEqual("(Next (Next (Next Proposition(phi))))", str(ast))
        
        # space is irrelevant (even when chained)
        ast = parsing.parseLTL("()()()phi")
        self.assertEqual("(Next (Next (Next Proposition(phi))))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("()TRUE")
        self.assertEqual("(Next Constant(TRUE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("()(a | b)")
        self.assertEqual("(Next (Proposition(a) Or Proposition(b)))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("()[]a")
        self.assertEqual("(Next (Globally Proposition(a)))", str(ast))
        
    def test_parse_until(self):
        ast = parsing.parseLTL("phi U psi")
        self.assertEqual("(Proposition(phi) Until Proposition(psi))", str(ast))

        # operator name is case sensitive (cause parsing to end)
        ast = parsing.parseLTL("phi u psi")
        self.assertEqual("Proposition(phi)", str(ast))
        
        # spacing is relevant (U is a keyword, not a literal -> space imposed)
        ast = parsing.parseLTL("phiUpsi")
        self.assertEqual("Proposition(phiUpsi)", str(ast))
        # can be chained (right associative)
        ast = parsing.parseLTL("phi U psi U chi")
        self.assertEqual("((Proposition(phi) Until Proposition(psi)) Until Proposition(chi))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("TRUE U FALSE")
        self.assertEqual("(Constant(TRUE) Until Constant(FALSE))", str(ast))
        ast = parsing.parseLTL("TRUE U alpha")
        self.assertEqual("(Constant(TRUE) Until Proposition(alpha))", str(ast))
        ast = parsing.parseLTL("alpha U FALSE")
        self.assertEqual("(Proposition(alpha) Until Constant(FALSE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("(a | b) U x")
        self.assertEqual("((Proposition(a) Or Proposition(b)) Until Proposition(x))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("[]a U <>b")
        self.assertEqual("((Globally Proposition(a)) Until (Eventually Proposition(b)))", str(ast))
        ast = parsing.parseLTL("[]a U b")
        self.assertEqual("((Globally Proposition(a)) Until Proposition(b))", str(ast))
        ast = parsing.parseLTL("a U <>b")
        self.assertEqual("(Proposition(a) Until (Eventually Proposition(b)))", str(ast))
        
    def test_parse_weak_until(self):
        ast = parsing.parseLTL("phi W psi")
        self.assertEqual("(Proposition(phi) WeakUntil Proposition(psi))", str(ast))

        # operator name is case sensitive (cause parsing to end)
        ast = parsing.parseLTL("phi w psi")
        self.assertEqual("Proposition(phi)", str(ast))
        
        # spacing is relevant (U is a keyword, not a literal -> space imposed)
        ast = parsing.parseLTL("phiWpsi")
        self.assertEqual("Proposition(phiWpsi)", str(ast))
        # can be chained (right associative)
        ast = parsing.parseLTL("phi W psi W chi")
        self.assertEqual("((Proposition(phi) WeakUntil Proposition(psi)) WeakUntil Proposition(chi))", str(ast))
        
        # it may accept a constant
        ast = parsing.parseLTL("TRUE W FALSE")
        self.assertEqual("(Constant(TRUE) WeakUntil Constant(FALSE))", str(ast))
        ast = parsing.parseLTL("TRUE W alpha")
        self.assertEqual("(Constant(TRUE) WeakUntil Proposition(alpha))", str(ast))
        ast = parsing.parseLTL("alpha W FALSE")
        self.assertEqual("(Proposition(alpha) WeakUntil Constant(FALSE))", str(ast))
        
        # it may accept a parenthesized expression
        ast = parsing.parseLTL("(a | b) W x")
        self.assertEqual("((Proposition(a) Or Proposition(b)) WeakUntil Proposition(x))", str(ast))
                
        # it may accept a timed expression
        ast = parsing.parseLTL("[]a W <>b")
        self.assertEqual("((Globally Proposition(a)) WeakUntil (Eventually Proposition(b)))", str(ast))
        ast = parsing.parseLTL("[]a W b")
        self.assertEqual("((Globally Proposition(a)) WeakUntil Proposition(b))", str(ast))
        ast = parsing.parseLTL("a W <>b")
        self.assertEqual("(Proposition(a) WeakUntil (Eventually Proposition(b)))", str(ast))
        
    def test_parse_parenthesized(self):
        # no impact on vars
        ast = parsing.parseLTL("(a)")
        self.assertEqual("Proposition(a)", str(ast))
        
        # it permits to change the side-associativity
        # can be chained (right associative)
        ast = parsing.parseLTL("(phi | psi) | chi")
        self.assertEqual("((Proposition(phi) Or Proposition(psi)) Or Proposition(chi))", str(ast))
        
        ast = parsing.parseLTL("(phi & psi) & chi")
        self.assertEqual("((Proposition(phi) And Proposition(psi)) And Proposition(chi))", str(ast))
        
        ast = parsing.parseLTL("(phi => psi) => chi")
        self.assertEqual("((Proposition(phi) Imply Proposition(psi)) Imply Proposition(chi))", str(ast))
        
        ast = parsing.parseLTL("(phi <=> psi) <=> chi")
        self.assertEqual("((Proposition(phi) Equiv Proposition(psi)) Equiv Proposition(chi))", str(ast))
        
        ast = parsing.parseLTL("([]a) & b")
        self.assertEqual("((Globally Proposition(a)) And Proposition(b))", str(ast))
        
        ast = parsing.parseLTL("(<>a) & b")
        self.assertEqual("((Eventually Proposition(a)) And Proposition(b))", str(ast))
        
        ast = parsing.parseLTL("([]a) U (<>b)")
        self.assertEqual("((Globally Proposition(a)) Until (Eventually Proposition(b)))", str(ast))
        
        ast = parsing.parseLTL("([]a) W (<>b)")
        self.assertEqual("((Globally Proposition(a)) WeakUntil (Eventually Proposition(b)))", str(ast))
        
    def test_precedence(self):
        ast = parsing.parseLTL("!a U b)")
        self.assertEqual("((Not Proposition(a)) Until Proposition(b))", str(ast))
        